freeslot("MT_CHAO")

freeslot(
    "S_CHAO_PICKACTIVITY",
	
	-- SPR_CHAI
    "S_CHAO_STAND",
	"S_CHAO_HELD",
	
	-- SPR_CHAD
	"S_CHAO_DANCE",
    
	-- SPR_CHAH
	"S_CHAO_JUMP",
    
	-- SPR_CHAU
	"S_CHAO_SIT",
	
	-- SPR_CHAW
	"S_CHAO_WALK_START",
	"S_CHAO_WALK",
	
	"S_CHAO_ORB"
)

freeslot(
    "SPR_CHAC",
	
	-- Dance
	"SPR_CHAD",
	
	-- Happy
	"SPR_CHAH",
	
	-- Idle
	"SPR_CHAI",
	
	-- Upset
	"SPR_CHAU",
	
	-- Walk
	"SPR_CHAW",
	
	-- Thingies above chao head
	
	-- Heart
    "SPR_CHHH",
	
	-- Regular orb
	"SPR_CHHO",
	
	-- Ping
	"SPR_CHHQ"
)

local function findClosestPlayer(mo, mindist)
	local closest
	
	for p in players.iterate do
		if p.spectator then return end
		
		local dist = R_PointToDist2(mo.x, mo.y, p.mo.x, p.mo.y)
		
		if dist < mindist then
			mindist = dist
			closest = p
		end
	end
	
	return closest
end

-- Returns the Z momentum of the object, accounting for slopes if the object is grounded
local function P_GetMobjZMovement(mo)
    local slope = mo.standingslope

    if not P_IsObjectOnGround(mo) then
        return mo.momz
    end

    if not slope then
        return 0
    end

    local angDiff = R_PointToAngle2(0, 0, mo.momx, mo.momy) - slope.xydirection
    local speed = FixedHypot(mo.momx, mo.momy)

    return P_ReturnThrustY(mo, slope.zangle, P_ReturnThrustX(mo, angDiff, speed))
end

local activities = {
	S_CHAO_DANCE,
	S_CHAO_JUMP,
	S_CHAO_SIT,
	
	-- Repeating those so chao is more likely to walk somewhere
	S_CHAO_WALK_START,
	S_CHAO_WALK_START,
	S_CHAO_WALK_START,
}

function A_ChaoPickActivity(mo)
	if not P_IsObjectOnGround(mo) then
		mo.state = S_CHAO_STAND
		return
	end

    local val = P_RandomRange(-5, #activities)

    if val > 0 then
		mo.state = activities[val]
    else
        mo.state = S_CHAO_STAND
    end
end

function A_ChaoPickWalkDest(mo, var1)
	if mo.spawn == nil then
		mo.spawn = { x = mo.x, y = mo.y }
	end
	
	mo.walk_target = {
		x = mo.spawn.x + P_RandomRange(-var1, var1)*mapobjectscale,
		y = mo.spawn.y + P_RandomRange(-var1, var1)*mapobjectscale,
	}
end

function A_ChaoLook(mo, var1)
	local lookp = findClosestPlayer(mo, var1*mapobjectscale)
	local look = lookp and lookp.mo
	
	if look then
		mo.angle = R_PointToAngle2(mo.x, mo.y, look.x, look.y)
	end
end

local function releaseChao(mo, pmo)
	mo.state = S_CHAO_STAND
	
	if pmo then
		local p = pmo.player

		mo.momx = pmo.momx
		mo.momy = pmo.momy
		mo.momz = P_GetMobjZMovement(pmo)
		
		local maxspeed = 4*FRACUNIT
		P_SetObjectMomZ(mo, FixedMul(4*FRACUNIT, FixedDiv(min(maxspeed, p.speed), maxspeed)), true)
	end
		
	-- Reset "spawn" position (used by chao walking)
	mo.spawn = nil
end

function A_ChaoHeld(mo)
	if not (mo.target and mo.target.player and mo.target.player.deadtimer == 0 and not mo.target.player.spectator) then
		releaseChao(mo)
		return
	end
	
	local pmo = mo.target
	local p = pmo.player
	
	if p.pickchao == 1 then
		releaseChao(mo, pmo)
		return
	end
	
	P_MoveOrigin(mo,
		pmo.x + P_ReturnThrustX(pmo, pmo.angle, pmo.radius),
		pmo.y + P_ReturnThrustY(pmo, pmo.angle, pmo.radius),
		pmo.z + 3*pmo.height/5)
	mo.angle = pmo.angle
	mo.momz = 0
end

function A_ChaoOrbFollow(mo, stiff, damp)
	if not mo.target then
		P_RemoveMobj(mo)
		return
	end
	
	local z = mo.target.z + mo.target.height + mapobjectscale*4
	
	-- x and y position is fixed for now
	P_MoveOrigin(mo, mo.target.x, mo.target.y, mo.z)
	
	local dist = z-mo.z
	
	if abs(dist) < mapobjectscale/10 then return end
	
	-- Time for Hooke's law!
	local force = FixedMul(stiff, dist)
	
	local damper_force = 0

	if mo.orb_lastforce ~= nil then
		damper_force = FixedMul(force - mo.orb_lastforce, damp)
	end

	mo.momz = $ + force + damper_force
	
	mo.orb_lastforce = force
end

local function P_AltFlip(n, tics, slow)
	slow = slow or 1

	if (leveltime/slow) % (2 * tics) < tics then
		return n
	end

	return -n
end

local function P_LerpFlip(n, tics, slow)
	slow = slow or 1

	local w = 2 * tics

	return P_AltFlip((((leveltime/slow) % w) - tics) * n, w, slow)
end

mobjinfo[MT_CHAO] = {
    spawnstate = S_CHAO_STAND,
    radius = 10*FRACUNIT,
    height = 25*FRACUNIT,
}

states[S_CHAO_PICKACTIVITY] = {
    tics = 0,
    action = A_ChaoPickActivity,
}

states[S_CHAO_STAND] = {
    sprite = SPR_CHAI,
    frame = A,
    tics = TICRATE*2,
	action = A_ChaoLook,
	var1 = 300,
    nextstate = S_CHAO_PICKACTIVITY,
}

states[S_CHAO_HELD] = {
    sprite = SPR_CHAI,
    frame = A,
    tics = 1,
	action = A_ChaoHeld,
    nextstate = S_CHAO_HELD,
}

states[S_CHAO_DANCE] = {
    sprite = SPR_CHAD,
    frame = A,
    tics = TICRATE*5,
    nextstate = S_CHAO_STAND,
}

states[S_CHAO_JUMP] = {
    sprite = SPR_CHAH,
    frame = A,
    tics = TICRATE*5,
    nextstate = S_CHAO_STAND,
}

states[S_CHAO_SIT] = {
    sprite = SPR_CHAU,
    frame = A|FF_ANIMATE,
    tics = TICRATE*10,
	var1 = 1,
	var2 = TICRATE/2,
    nextstate = S_CHAO_STAND,
}

states[S_CHAO_WALK_START] = {
	tics = 0,
	action = A_ChaoPickWalkDest,
	var1 = 400,
	nextstate = S_CHAO_WALK,
}

states[S_CHAO_WALK] = {
    sprite = SPR_CHAW,
    frame = A,
    tics = -1,
    nextstate = S_CHAO_WALK,
}

states[S_CHAO_ORB] = {
	sprite = SPR_CHHO,
	frame = A,
	tics = 1,
	action = A_ChaoOrbFollow,
	var1 = 2*FRACUNIT/3,
	var2 = 2*FRACUNIT/3,
	nextstate = S_CHAO_ORB,
}

addHook("MobjThinker", function(mo)
    if not mo.valid then return end -- :))))))))))))
	
	-- Stop any momentum on ground
	if P_IsObjectOnGround(mo) then
		mo.momx = 0
		mo.momy = 0
	end

    if mo.state == S_CHAO_JUMP and P_IsObjectOnGround(mo) then
        P_SetObjectMomZ(mo, 8*FRACUNIT, false)
    elseif mo.state == S_CHAO_DANCE then
		mo.frame = 1 + P_LerpFlip(1, 1, 8)
	elseif mo.state == S_CHAO_WALK then
		mo.frame = 1 + P_LerpFlip(1, 1, TICRATE/6)
	
		if R_PointToDist2(mo.x, mo.y, mo.walk_target.x, mo.walk_target.y) <= mapobjectscale then
			mo.state = S_CHAO_STAND
		else
			mo.angle = R_PointToAngle2(mo.x, mo.y, mo.walk_target.x, mo.walk_target.y)
			
			local speed = 2*FRACUNIT
			if not P_TryMove(mo, mo.x + P_ReturnThrustX(mo, mo.angle, speed), mo.y + P_ReturnThrustY(mo, mo.angle, speed), false) then
				mo.state = S_CHAO_STAND
			end
		end
	elseif mo.state == S_CHAO_STAND then
		local closest = findClosestPlayer(mo, 60*mapobjectscale)
		
		if closest and closest.pickchao == 1 then
			closest.pickchao = $ + 1 -- Hack to not release chao immediately after picking up
			mo.target = closest.mo
			mo.state = S_CHAO_HELD
		end
	end
end, MT_CHAO)

addHook("PlayerThink", function(p)
	if p.spectator then return end
	
	if p.cmd.buttons & BT_CUSTOM2 then
		p.pickchao = ($ or 0) + 1
	else
		p.pickchao = 0
	end
end)

local function spawnChao(x, y, z)
    local chao = P_SpawnMobj(x, y, z, MT_CHAO)
	chao.scale = 6*mapobjectscale/5
    
	P_SpawnShadowMobj(chao)
	
	local orb = P_SpawnMobj(x, y, z+chao.height, MT_THOK)
	orb.target = chao
	orb.state = S_CHAO_ORB
	
    return chao
end

COM_AddCommand("chao", function(p)
    local chao = spawnChao(p.mo.x, p.mo.y, p.mo.z + mapobjectscale*10)
    chao.color = P_RandomRange(1, MAXSKINCOLORS-1)
    chao.angle = p.mo.angle
end, COM_ADMIN)
