-- Fixed-point constants
local FRACUNIT = 0x10000
local FLOATSPEED = 0x1800 -- Set to appropriate value
local DI_NODIR = 8
local NUMDIRS = 8

-- Precomputed fixed-point movement values
local xspeed = {
    FRACUNIT,
    46341,
    0,
    -46341,
    -FRACUNIT,
    -46341,
    0,
    46341
}

local yspeed = {
    0,
    46341,
    FRACUNIT,
    46341,
    0,
    -46341,
    -FRACUNIT,
    -46341
}

local FRACUNIT = 0x10000
local FLOATSPEED = 0x1800
local DI_NODIR = 8
local NUMDIRS = 8

local MF_DROPOFF = 0x4000000
local MF_FLOAT = 0x00001000
local MF2_TWOD = 0x00000001
local MF2_INFLOAT = 0x00000002

local function P_MoveDrop(actor, speed)
    if not actor or not actor.valid then return false end

    local movedir = actor.movedir

    local scale = actor.scale or FRACUNIT
    local xspd = xspeed[movedir + 1] or 0
    local yspd = yspeed[movedir + 1] or 0

    local tryx = actor.x + FixedMul(speed * xspd, scale)
    local tryy = (twodlevel or (actor.flags2 & MF2_TWOD) ~= 0)
        and actor.y
        or actor.y + FixedMul(speed * yspd, scale)

    local originalFlags = actor.flags
    actor.flags = $ | MF_DROPOFF

    local success = P_TryMove(actor, tryx, tryy, true)

    if not success then
        -- Retry with noclip enabled temporarily
        success = P_TryMove(actor, tryx, tryy, true)
    end

    actor.flags = originalFlags -- Restore original flags (removes NOCLIP and DROPOFF)

    -- Floating enemies can adjust Z even if stuck
    if not success and (actor.flags & MF_FLOAT) ~= 0 and floatok then
        local zchange = FixedMul(FLOATSPEED, scale)
        if actor.z < tmfloorz then
            actor.z = actor.z + zchange
        else
            actor.z = actor.z - zchange
        end

        if actor.type == MT_JETJAW and actor.z + actor.height > actor.watertop then
            actor.z = actor.watertop - actor.height
        end

        actor.flags2 = $ | MF2_INFLOAT
        return true
    else
        actor.flags2 = $ & ~MF2_INFLOAT
    end

    if not success then
        P_NewChaseDir(actor)
    end

    return success
end

function A_ChaseDropOff(actor, var1, var2)
    if not actor or not actor.valid then return end

    local skipMelee     = (var1 & 1) ~= 0
    local skipMissile   = (var1 & 2) ~= 0

    -- Reduce reaction time
    if actor.reactiontime and actor.reactiontime > 0 then
        actor.reactiontime = $ - 1
    end

    -- Reduce threshold if applicable
    if actor.threshold and actor.threshold > 0 then
        if not actor.target or not actor.target.valid or actor.target.health <= 0 then
            actor.threshold = 0
        else
            actor.threshold = $ - 1
        end
    end

    -- Adjust angle to match movement direction
    if actor.movedir and actor.movedir < 8 then
        local idealangle = actor.movedir * ANGLE_45
        local delta = actor.angle - idealangle

        if delta > 0 then
            actor.angle = $ - ANGLE_45
        elseif delta < 0 then
            actor.angle = $ + ANGLE_45
        end
    end

    -- No valid target? Try finding a new one
    if not actor.target or not actor.target.valid or not (actor.target.flags & MF_SHOOTABLE) then
        if P_LookForPlayers(actor, 99999, false, 0) then
            return
        end

        actor.state = actor.info.spawnstate
        return
    end

    -- Skip if just attacked
    if actor.flags2 & MF2_JUSTATTACKED then
        actor.flags2 = $ & ~MF2_JUSTATTACKED
        P_NewChaseDir(actor)
        return
    end

    -- Melee attack
    if not skipMelee and actor.info.meleestate and P_CheckMeleeRange(actor) then
        if actor.info.attacksound then
            S_StartSound(actor, actor.info.attacksound)
        end
		
        actor.state = actor.info.meleestate
        return
    end

    -- Missile attack
    if not skipMissile and actor.info.missilestate and (not actor.movecount or actor.movecount == 0) and P_CheckMissileRange(actor) then
        actor.state = actor.info.missilestate
        actor.flags2 = $ | MF2_JUSTATTACKED
        return
    end

    -- Multiplayer re-targeting
    if multiplayer and actor.threshold == 0
    and (not actor.target.valid or actor.target.health <= 0 or not P_CheckSight(actor, actor.target)) then
        if P_LookForPlayers(actor, true, false, 0) then
            return
        end
    end

    -- Movement toward player
    actor.movecount = $ - 1
    if actor.movecount < 0 or not P_MoveDrop(actor, actor.info.speed) then
        P_NewChaseDir(actor)
    end
end