freeslot("MT_MADFATHER", "MT_BLOOD")
freeslot("sfx_madfo1")
freeslot("sfx_madfo2")
freeslot("sfx_madfo3")
freeslot("sfx_madfo4")
freeslot("sfx_madch1")
freeslot("sfx_madch2")
freeslot("sfx_madid1")
freeslot("sfx_madid2")
freeslot("sfx_madid3")
freeslot("sfx_madid4")
freeslot("sfx_madhu1")
freeslot("sfx_madhu2")

local cv_madfatherjumpscare = CV_RegisterVar({
  name = "sg_madfatherjumpscare",
  defaultvalue = "Off",
  possiblevalue = CV_OnOff,
})

local madfatheridle = {
    [0] = sfx_madid1, 
    sfx_madid2, 
    sfx_madid3, 
    sfx_madid4
}

local madfatherchase = {
    [0] = sfx_madfo1, 
    sfx_madfo2, 
    sfx_madfo3, 
    sfx_madfo4
}

function A_BunnyHopOrSwim(actor, var1, var2)
    if P_IsObjectOnGround(actor) then
        P_SetObjectMomZ(actor, var1*FRACUNIT, false)
        P_InstaThrust(actor, actor.angle, FixedMul(var2*FRACUNIT, actor.scale))
    elseif actor.waterfloat then
		P_SetObjectMomZ(actor, var1*FRACUNIT*1, false)
        P_InstaThrust(actor, actor.angle, FixedMul(var2*FRACUNIT, actor.scale))
    end
end

local INVINCIBILITY_DURATION = 8 * TICRATE

addHook("MobjSpawn", function(mo)
  mo.scale = mapobjectscale*2
  mo.destscale = mo.scale
end, MT_MADFATHER)

addHook("MobjThinker", function(mo)
    if not mo.valid then return end

    -- Reduce the rain timer, if it exists
    mo.soundtimer = max(($ or 0) - 1, 0)
	mo.chainsawsoundtimer = max(($ or 0) - 1, 0)

	local idleStates = {
		[S_MADF_IDLE0] = true,
		[S_MADF_IDLE1] = true,
		[S_MADF_IDLE2] = true,
		[S_MADF_IDLE3] = true,
		[S_MADF_IDLE4] = true,
		[S_MADF_IDLE5] = true,
		[S_MADF_IDLE6] = true,
		[S_MADF_ROAMING1] = true,
		[S_MADF_ROAMING2] = true,
		[S_MADF_ROAMING3] = true,
		[S_MADF_ROAMING4] = true,
	}
	
	local chaseStates = {
		[S_MADF_CHASE1] = true,
		[S_MADF_CHASE2] = true,
		[S_MADF_CHASE3] = true,
		[S_MADF_CHASE4] = true,
	}
	
	if mo.chaselooptimer == 0 then
		S_StartSound(mo, sfx_madch1)
		S_StartSound(mo, sfx_madch1)
		mo.chaselooptimer = 54
	end
	
    if idleStates[mo.state] then
        -- If idle and rain timer has reached 0, play a random sound from madfathersfx
        if mo.soundtimer == 0 then
            local sound = madfatheridle[P_RandomKey(#madfatheridle)]
            S_StartSound(mo, sound)
			S_StartSound(mo, sound)
            mo.soundtimer = 10 * TICRATE -- Reset the timer to play the sound again after 10 seconds
        end
    end
	
    if chaseStates[mo.state] then
        -- If idle and rain timer has reached 0, play a random sound from madfathersfx
        if mo.soundtimer == 0 then
            local sound = madfatherchase[P_RandomKey(#madfatherchase)]
            S_StartSound(mo, sound)
			S_StartSound(mo, sound)
            mo.soundtimer = 10 * TICRATE -- Reset the timer to play the sound again after 10 seconds
        end
    end
	
	-- Check if actor is in invincibility frames
    if mo.invincible then
        mo.invincibilityTimer = mo.invincibilityTimer - 1

        -- End invincibility after duration expires
        if mo.invincibilityTimer <= 0 then
            mo.invincible = false
        end
    end
	
	if mo.health == 1 and mo.flags & MF_ENEMY then
		S_StartSound(nil, sfx_madhu2)
		mo.state = S_MADF_DEATH1
		mo.flags = MF_SOLID
		P_LinedefExecute(7551)
	end
end, MT_MADFATHER)

addHook("MobjCollide", function(mo, projectile)
	if not mo.valid then return end
	if not projectile.valid then return end
	if not projectile then return end
	if mo.z > projectile.z + projectile.height
		or projectile.z > mo.z +mo.height then

		return
	end
	
	if mo.health == 1 then return end
	
    -- If actor isn't invincible, apply damage and set invincibility
    if not mo.invincible and (projectile.type == MT_ORBINAUT or projectile.type == MT_JAWZ or projectile.type == MT_ORBINAUT_SHIELD or projectile.type == MT_BANANA or projectile.type == MT_SHOTGUNBULLET) and not projectile.player then
        mo.health = $ - 1

        -- Set invincibility
        mo.invincible = true
        mo.invincibilityTimer = INVINCIBILITY_DURATION

        -- Play hurt sound when damaged
        S_StartSound(mo, sfx_madhu1)
		mo.state = S_MADF_HURT1
    end
end, MT_MADFATHER)

addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
    -- Check if the inflictor is a MineExplosion or SPBExplosion
    if inflictor.type == MT_MINEEXPLOSION or inflictor.type == MT_SPBEXPLOSION then
        if inflictor.extravalue1 == 1 then
			return false
		end
		-- Prevent damage from these explosions
        return false
    end
    
    -- Allow default behavior for all other damage types
    return true
end, MT_MADFATHER)

addHook("TouchSpecial", function(mo, toucher)
    if not toucher.player then return true end -- Ensure toucher is a player

    local player = toucher.player
	
	if player.kartstuff[k_respawn] or player.powers[pw_flashing] then return true end

    -- Freeze player movement
    player.mo.momx = 0
    player.mo.momy = 0
    player.mo.momz = 0
    player.pflags = $|PF_FULLSTASIS -- Freezes the player

    -- Optional: Prevent further control
    player.powers[pw_nocontrol] = TICRATE
	
	if player._madfather_caught then return true end
	
    -- Delay before death and jumpscare
    player.playerstate = PST_LIVE -- Stay alive briefly
	player.deathTimer = 2*TICRATE

	player.jumpscareTimer = 1
	player.jumpscareY = 200
	
	S_StartSound(toucher, sfx_madfo4)
	player._madfather_caught = true

    return true -- Prevent object from disappearing
end, MT_MADFATHER)

addHook("PlayerThink", function(player)
	if not player.mo then return end
	
	-- Animate jumpscare
	if player.jumpscareTimer then
		player.jumpscareTimer = $ + 1
		if player.jumpscareY and player.jumpscareY > 1 then
			player.jumpscareY = $ - 12
		end
	end

	-- Handle death
	if player.deathTimer then
		player.deathTimer = $ - 1
		
		for i = 1, 5 do
			local x = player.mo.x
			local y = player.mo.y
			local z = player.mo.z

			local blood = P_SpawnMobj(x, y, z, MT_BLOOD) -- Replace with your blood object if needed
			blood.scale = FRACUNIT/2 -- Smaller size

			-- Small outward random push
			blood.momx = P_RandomRange(-5, 5)*FRACUNIT/10
			blood.momy = P_RandomRange(-5, 5)*FRACUNIT/10

			-- Upward fountain burst
			blood.momz = FRACUNIT + P_RandomRange(1, 5)*FRACUNIT
		end
		
		if player.deathTimer <= 0 then
			P_KillMobj(player.mo)
			player.deathTimer = nil
			player._madfather_caught = false
		end
	end

	-- Reset "caught" flag if no longer dead or invulnerable
	if player.playerstate ~= PST_DEAD and not player.deathTimer then
		player._madfather_caught = false
	end
end)

hud.add(function(v, p)
	if not p or not p._madfather_caught then return end

	if cv_madfatherjumpscare.value then
		local patch = v.cachePatch("MADJUMP")
		local y = p.jumpscareY or 200
		v.draw(0, y, patch, V_HUDTRANS | V_SNAPTOTOP)
	end
end)